from enum import Enum


class QualityMode(str, Enum):
    Benchmark = "BENCHMARK"
    Consensus = "CONSENSUS"


BENCHMARK_AUTO_AUDIT_NUMBER_OF_LABELS = 1
BENCHMARK_AUTO_AUDIT_PERCENTAGE = 1
CONSENSUS_AUTO_AUDIT_NUMBER_OF_LABELS = 3
CONSENSUS_AUTO_AUDIT_PERCENTAGE = 0
